@@ -1,4 +1,4 @@
-#!/usr/pkg/bin/python3.11
+#!/usr/pkg/bin/python3.13
 
 ##===--- iwyu_tool.py -----------------------------------------------------===##
 #
@@ -35,6 +35,7 @@ import sys
 import json
 import time
 import shlex
+import shutil
 import argparse
 import tempfile
 import subprocess
@@ -52,7 +53,7 @@ LINES_RE = re.compile(r'^- (.*?)  // lines ([0-9]+)-[0-9]+$')
 GENERAL, ADD, REMOVE, LIST = range(4)
 
 
-def clang_formatter(output):
+def clang_formatter(output, style):
     """ Process iwyu's output into something clang-like. """
     formatted = []
 
@@ -83,15 +84,19 @@ def clang_formatter(output):
         elif state[0] == ADD:
             match = ADD_RE.match(line)
             if match:
-                formatted.append("%s:1:1: error: add '%s' (%s)" %
-                                 (state[1], match.group(1), match.group(2)))
+                formatted.append("%s:1:1: %s: add '%s' (%s)" %
+                                 (state[1],
+                                  style,
+                                  match.group(1),
+                                  match.group(2)))
             else:
-                formatted.append("%s:1:1: error: add '%s'" % (state[1], line))
+                formatted.append("%s:1:1: %s: add '%s'" %
+                                 (state[1], style, line))
         elif state[0] == REMOVE:
             match = LINES_RE.match(line)
             line_no = match.group(2) if match else '1'
-            formatted.append("%s:%s:1: error: superfluous '%s'" %
-                             (state[1], line_no, match.group(1)))
+            formatted.append("%s:%s:1: %s: superfluous '%s'" %
+                             (state[1], line_no, style, match.group(1)))
 
     return os.linesep.join(formatted)
 
@@ -99,7 +104,8 @@ def clang_formatter(output):
 DEFAULT_FORMAT = 'iwyu'
 FORMATTERS = {
     'iwyu': lambda output: output,
-    'clang': clang_formatter
+    'clang': lambda output: clang_formatter(output, style="error"),
+    'clang-warning': lambda output: clang_formatter(output, style="warning"),
 }
 
 
@@ -179,7 +185,7 @@ def win_split(cmdline):
             elif c in (' ', '\t') and not in_quotes:
                 # MSDN: Arguments are delimited by white space, which is either
                 # a space or a tab [but only outside of a string].
-                # Flush backslashes and return arg bufferd so far, unless empty.
+                # Flush any buffered backslashes and yield arg, unless empty.
                 arg += '\\' * backslashes
                 if arg:
                     yield arg
@@ -212,21 +218,20 @@ def split_command(cmdstr):
 
 def find_include_what_you_use():
     """ Find IWYU executable and return its full pathname. """
-    if 'IWYU_BINARY' in os.environ:
-        return os.environ.get('IWYU_BINARY')
+    env_iwyu_path = os.environ.get('IWYU_BINARY')
+    if env_iwyu_path:
+        return os.path.realpath(env_iwyu_path)
 
-    # TODO: Investigate using shutil.which when Python 2 has passed away.
-    executable_name = 'include-what-you-use'
-    if sys.platform.startswith('win'):
-        executable_name += '.exe'
+    # Search in same dir as this script.
+    iwyu_path = shutil.which('include-what-you-use',
+                             path=os.path.dirname(__file__))
+    if iwyu_path:
+        return os.path.realpath(iwyu_path)
 
-    search_path = [os.path.dirname(__file__)]
-    search_path += os.environ.get('PATH', '').split(os.pathsep)
-
-    for dirpath in search_path:
-        full = os.path.join(dirpath, executable_name)
-        if os.path.isfile(full):
-            return os.path.realpath(full)
+    # Search the system PATH.
+    iwyu_path = shutil.which('include-what-you-use')
+    if iwyu_path:
+        return os.path.realpath(iwyu_path)
 
     return None
 
@@ -360,6 +365,26 @@ def slice_compilation_db(compilation_db, selection):
     return new_db
 
 
+def worst_exit_code(worst, cur):
+    """Return the most extreme exit code of two.
+
+    Negative exit codes occur if the program exits due to a signal (Unix) or
+    structured exception (Windows). If we've seen a negative one before, keep
+    it, as it usually indicates a critical error.
+
+    Otherwise return the biggest positive exit code.
+    """
+    if cur < 0:
+        # Negative results take precedence, return the minimum
+        return min(worst, cur)
+    elif worst < 0:
+        # We know cur is non-negative, negative worst must be minimum
+        return worst
+    else:
+        # We know neither are negative, return the maximum
+        return max(worst, cur)
+
+
 def execute(invocations, verbose, formatter, jobs, max_load_average=0):
     """ Launch processes described by invocations. """
     exit_code = 0
@@ -367,7 +392,7 @@ def execute(invocations, verbose, formatter, jobs, max_load_average=0):
         for invocation in invocations:
             proc = invocation.start(verbose)
             print(formatter(proc.get_output()))
-            exit_code = max(exit_code, proc.returncode)
+            exit_code = worst_exit_code(exit_code, proc.returncode)
         return exit_code
 
     pending = []
@@ -377,7 +402,7 @@ def execute(invocations, verbose, formatter, jobs, max_load_average=0):
         for proc in complete:
             pending.remove(proc)
             print(formatter(proc.get_output()))
-            exit_code = max(exit_code, proc.returncode)
+            exit_code = worst_exit_code(exit_code, proc.returncode)
 
         # Schedule new processes if there's room.
         capacity = jobs - len(pending)
@@ -472,9 +497,13 @@ def _bootstrap(sys_argv):
                         choices=FORMATTERS.keys(), default=DEFAULT_FORMAT,
                         help='Output format (default: %s)' % DEFAULT_FORMAT)
     parser.add_argument('-j', '--jobs', type=int, default=1,
-                        help='Number of concurrent subprocesses')
+                        nargs='?', const=0,
+                        help=('Number of concurrent subprocesses. If zero, '
+                              'will try to match the logical cores of the '
+                              'system.'))
     parser.add_argument('-l', '--load', type=float, default=0,
-                        help='Do not start new jobs if the 1min load average is greater than the provided value')
+                        help=('Do not start new jobs if the 1min load average '
+                              'is greater than the provided value'))
     parser.add_argument('-p', metavar='<build-path>', required=True,
                         help='Compilation database path', dest='dbpath')
     parser.add_argument('source', nargs='*',
@@ -492,8 +521,12 @@ def _bootstrap(sys_argv):
     argv, extra_args = partition_args(sys_argv[1:])
     args = parser.parse_args(argv)
 
+    jobs = args.jobs
+    if jobs == 0:
+        jobs = os.cpu_count() or 1
+
     return main(args.dbpath, args.source, args.verbose,
-                FORMATTERS[args.output_format], args.jobs, args.load, extra_args)
+                FORMATTERS[args.output_format], jobs, args.load, extra_args)
 
 
 if __name__ == '__main__':
